package iotcloud;
import java.io.*;
import java.net.*;
import java.util.Arrays;
import javax.crypto.*;
import javax.crypto.spec.*;
import java.security.SecureRandom;

class CloudComm {
	String baseurl;
	Cipher encryptcipher;
	Cipher decryptcipher;
	Mac mac;
	byte[] salt;
	SecretKeySpec key;
	static final int SALT_SIZE = 8;

	
	CloudComm() {
	}

	CloudComm(String _baseurl, String password) {
		this.baseurl=_baseurl;
		initCloud(password);
	}

	private void initKey(String password) {
		try {
			salt=new byte[SALT_SIZE];
			SecureRandom random = new SecureRandom();
			random.nextBytes(salt);
			PBEKeySpec keyspec = new PBEKeySpec(password.toCharArray(), salt, 65536, 128);
			SecretKey key = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA256").generateSecret(keyspec);
			key = new SecretKeySpec(key.getEncoded(), "AES");
		} catch (Exception e) {
			e.printStackTrace();
			throw new Error("Failed generating key.");
		}
	}

	private void initCloud(String password) {
		try {
			initKey(password);
			mac = Mac.getInstance("HmacSHA256");
			mac.init(key);
		} catch (Exception e) {
			e.printStackTrace();
			throw new Error("Failed To Initialize Ciphers");
		}
	}
	
	private URL buildRequest(boolean isput, long sequencenumber, long maxentries) throws IOException {
		String reqstring=isput?"req=putslot":"req=getslot";
		String urlstr=baseurl+"?"+reqstring+"&seq="+sequencenumber;
		if (maxentries != 0)
			urlstr += "&max="+maxentries;
		return new URL(urlstr);
	}

	public Slot[] putSlot(Slot slot, int max) {
		try {
			long sequencenumber=slot.getSequenceNumber();
			byte[] bytes=slot.encode(mac);

			URL url=buildRequest(true, sequencenumber, max);
			URLConnection con=url.openConnection();
			HttpURLConnection http = (HttpURLConnection) con;
			http.setRequestMethod("POST");
			http.setFixedLengthStreamingMode(bytes.length + SALT_SIZE);
			http.setDoOutput(true);
			http.connect();
			OutputStream os=http.getOutputStream();
			os.write(salt);
			os.write(bytes);
			System.out.println(http.getResponseMessage());

			InputStream is=http.getInputStream();
			DataInputStream dis=new DataInputStream(is);
			byte[] resptype=new byte[7];
			dis.readFully(resptype);
			if (Arrays.equals(resptype, "getslot".getBytes()))
				return processSlots(dis);
			else if (Arrays.equals(resptype, "putslot".getBytes()))
				return null;
			else
				throw new Error("Bad response to putslot");
		} catch (Exception e) {
			throw new Error("putSlot failed");
		}
	}

	/*
			Cipher encryptCipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
			encryptCipher.init(Cipher.ENCRYPT_MODE, secret);
			Cipher decryptCipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
			decryptCipher.init(Cipher.DECRYPT_MODE, secret);
	*/
	
	public Slot[] getSlots(long sequencenumber) {
		try {
			URL url=buildRequest(false, sequencenumber, 0);
			URLConnection con=url.openConnection();
			HttpURLConnection http = (HttpURLConnection) con;
			http.setRequestMethod("POST");
			http.connect();
			System.out.println(http.getResponseMessage());
			InputStream is=http.getInputStream();

			DataInputStream dis=new DataInputStream(is);
			byte[] resptype=new byte[7];
			dis.readFully(resptype);
			if (!Arrays.equals(resptype, "getslot".getBytes()))
				throw new Error("Bad Response: "+new String(resptype));
			else
				return processSlots(dis);
		} catch (Exception e) {
			throw new Error("getSlots failed");
		}
	}

	Slot[] processSlots(DataInputStream dis) throws IOException {
		int numberofslots=dis.readInt();
		int[] sizesofslots=new int[numberofslots];
		Slot[] slots=new Slot[numberofslots];
		for(int i=0; i<numberofslots; i++)
			sizesofslots[i]=dis.readInt();

		for(int i=0; i<numberofslots; i++) {
			byte[] data=new byte[sizesofslots[i]];
			dis.readFully(data);
			slots[i]=Slot.decode(data, mac);
		}
		dis.close();
		return slots;
	}
}
